import java.io.*; /* for file input and output */

/** Bas2Prot.java
 * <P>
 * Converts a file of bases to a flat file database of proteins.
 * <P>
 * Copyright 04/09/97 George Ruban, Joachim Reidl
 * <P>
 * This software is in the public domain, and is freely available for any
 * use, with no guarantee, on an "as is" basis. The authors would be
 * interested in the ways in which this software is used and any problems
 * that may be encountered. Users are requested, though not required, to
 * send mail to the below addresses.
 * <P>
 * This program and a paper describing it, "A THEORETICAL APPROACH TO
 * IDENTIFY PUTATIVE SIGNAL SEQUENCES: PROVIDING EVIDENCE FOR SECRETED GENE
 * PRODUCTS BY PURE COMPUTATIONAL ANALYSIS," are available on the World
 * Wide Web.
 * <UL>
 * <LI>paper:    http://www.geocities.com/SiliconValley/Vista/2013/joachim.html
 * <LI>C source: http://www.geocities.com/SiliconValley/Vista/2013/bas2prot.txt
 * <LI>Javadoc HTML: http://www.geocities.com/SiliconValley/Vista/2013/b2p.html
 * <LI>Java source:  http://www.geocities.com/SiliconValley/Vista/2013/b2p.java
 * </UL>
 * <P>
 * Written (in C) 02/07/1995 by George Ruban
 * 67 Thurston Street, Somerville, MA 02145, USA.
 * http://www.geocities.com/SiliconValley/Vista/2013
 * email: gmn@csa.bu.edu
 * <P>
 * For Joachim Reidl
 * Research Center for Infectious Disease, University of Wuerzburg,
 * Roentgenring 11, 97070 Wuerzburg, Germany. Tel.: 0049 931 59509,
 * e-mail: joachim.reidl@rzroe.uni-wuerzburg.de
 * <P>
 * @version 1.1 04/09/1997 - changed from C++ style comments, published on World Wide Web
 * @version 2.0 03/31/1998 - replaced vonHeijne 1988 matrices with 1997 matrices,
 *      gave choice of 3 matrices to use
 * @version 2.1 04/19/1998 - converting to Java 
*/
public final class Bas2Prot implements Runnable 
// final for speed, very important with 
// huge data input expected
{
  /** Maximum length of a line of output for the protein database.
   * @see  #writeProtein  */
  public static final int MAXLINE=60;

  /** Maximum amino acids expected in a protein.  There are 4^3 = 64
   * different possible base triplets.  3 of them are stop codons. A length
   * of 8000 amino acids should be plenty for a single protein, I
   * hope... */
  public static final int MAXPROTEIN=8000;

  /** Maximum weight of a Eukaryotic protein, from table. */
  public static final int MAXEUKARYOTIC= 4087;
  /** Maximum weight of a Gram-Negative protein, from table. */
  public static final int MAXGRAMNEGATIVE= 1196;
  /** Maximum weight of a Gram-Positive protein, from table. */
  public static final int MAXGRAMPOSITIVE= 627;

  /** 6 arrays to hold amino acids, one per frame. */
  byte[][] buffer = new byte[6][MAXPROTEIN];

  /** Indices of current location in each frame buffer.  The first, forward
   * 3 point at the next to be filled, and increment - the last, backward,
   * 3 index one beyond the next to be filled, and decrement.
   * @see #buffer */
  int[] buffptr = { 0, 0, 0, MAXPROTEIN, MAXPROTEIN, MAXPROTEIN};
  
  /** Buffer to hold forward bases, for frames 0-2 */
  byte[] fbuf = {0, 0, 0, 0};
  /** Buffer to hold reverse bases, for frames 3-5 */
  byte[] rbuf = {0, 0, 0, 0}; 


  /** Minimum weight to write a protein. */
  public int minWeight = 0; 
  /* Maximum number of times to slide weighing window. */
  public int maxSlide = 0; 
  /** Minimum length of a protein in order to write it. */
  public int minLength = 0; 
  /** Maximum distance of K or R from starting M for a valid protein. */
  public int maxPositiveResidues = 0; 
  /** Number of bases read. */
  public int bases = 0;
  /** Number of proteins written. */
  public int proteins = 0;

  /** Number of amino acids are be excluded from protein output. */
  public int nExcluded;
  /** True to exclude matching alphabetical amino acid letter - that is, if
   * presence of corresponding amino acid means entire protein is to be
   * excluded. */
  public boolean[] exclude = new boolean[26]; 

  /** Input file/stream of bases. Non-bases ignored. */
  public InputStream in;

  /** Output file/stream of proteins. */
  public PrintStream out;

  /** Amino acid counts for eukaryotic signal sequences from Gunnar von
   * Heijne's paper in "Nucleic Acids Research" Volume 14, Number 11, 1986,
   * p.4683, IRL Press, Oxford.  3 x 26 x 15. 1 2 3 ... across the top, A-Y
   * across the side */
  static final int[][][] cleavage =
  {
  /* Eukaryotic cleavage sites */
  {
    /*      -13 -12 -11 -10  -9  -8  -7  -6  -5  -4  -3  -2  -1   1   2  */
    /* A */{101,112,106,100,158,128,107,149,146,107,258, 80,458,141, 55},
    /* no B's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* C */{ 36, 32, 34, 31, 42, 66, 57, 39, 30, 34, 68, 21, 50, 23, 27},
    /* D */{  2,  0,  3,  0,  3,  2,  4,  5, 25, 14,  4, 44,  6, 68, 66},
    /* E */{  1,  2,  4,  5,  4,  6, 13,  9, 28, 27,  7, 66,  6, 92, 88},
    /* F */{ 75, 67, 80, 81, 65, 58, 92, 67, 25, 34,  7, 46,  6, 34, 28},
    /* G */{ 39, 24, 26, 34, 36, 50, 40, 29,108,125, 74, 38,184, 52, 57},
    /* H */{  5,  3,  4,  2,  4, 10, 14,  7, 23, 12,  5, 53,  2, 23, 22},
    /* I */{ 73, 67, 82, 59, 52, 52, 47, 69, 22, 43, 41, 18,  4, 37, 34},
    /* no J's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* K */{  4,  1,  1,  2,  0,  3,  3,  3, 20, 19,  7, 24,  4, 52, 40},
    /* L */{376,432,397,449,386,329,333,280, 98,147, 65,165, 21, 60, 51},
    /* M */{ 26, 15, 21, 19, 11, 30, 20, 11,  5, 12,  9, 14,  3, 14, 12},
    /* N */{  2,  4,  4,  6,  7,  7,  8,  7, 25, 14,  9, 37,  9, 26, 53},
    /* no O's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* P */{ 23, 13, 17,  9, 12, 13, 24, 60, 98, 74,  7, 17, 27, 11,166},
    /* Q */{  3,  6,  4,  9, 11, 19, 18, 13, 55, 40,  5, 72, 15,135, 42},
    /* R */{  7,  4,  4,  2,  3,  7,  9,  6, 34, 38,  5, 56, 20, 33, 40},
    /* S */{ 57, 47, 40, 44, 65, 76, 67, 66,117, 97,142,112,141, 91, 88},
    /* T */{ 32, 39, 30, 38, 38, 49, 42, 35, 70, 70,103, 44, 43, 39, 57},
    /* no U's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* V */{112,124,127,100, 99, 86, 79,129, 60, 72,191, 45,  8, 50, 56},
    /* W */{ 16, 13, 20, 11,  9, 13, 24, 21, 13, 13,  2, 33,  2,  9,  6},
    /* no X's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* Y */{ 17,  3,  6,  9,  6,  7, 10,  6,  8, 19,  2, 26,  2, 21, 23},
    /* no Z's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  },
  /* Gram-negative cleavage sites */
  {
    /*      -13 -12 -11 -10  -9  -8  -7  -6  -5  -4  -3  -2  -1   1   2 */
    /* A */{ 63, 58, 56, 64, 63, 39, 41, 78, 34, 41,156, 20,229,102, 19},
    /* no B's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* C */{ 10,  6,  5,  5,  6,  5,  4,  7,  3,  0,  3,  3,  0,  4,  0},
    /* D */{  0,  1,  1,  2,  1,  2,  2,  0,  0,  1,  2,  2,  0, 17, 43},
    /* E */{  0,  0,  0,  1,  1,  0,  1,  0,  1,  2,  1,  4,  1, 21, 40},
    /* F */{ 15, 12, 18, 12, 17, 32, 31,  7, 17, 13,  1, 28,  0,  2,  1},
    /* G */{ 20, 22, 17, 13, 11, 29, 19, 13, 41, 15,  5,  5, 11, 12, 15},
    /* H */{  1,  0,  0,  0,  1,  0,  2,  1,  7,  1,  1, 32,  0,  2,  0},
    /* I */{ 17, 22, 16, 15, 16,  5, 25, 11,  4,  9,  3,  5,  0,  3,  9},
    /* no J's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* K */{  0,  1,  2,  0,  0,  0,  1,  0,  1,  4,  1,  3,  0, 12,  4},
    /* L */{ 60, 61, 71, 63, 67, 82, 44,  7, 31, 12, 15, 42,  4,  8, 11},
    /* M */{ 11,  9,  9,  8, 17, 12, 17,  9,  4,  2,  0, 19,  0,  3,  2},
    /* N */{  2,  1,  2,  2,  3,  0,  2,  5,  5, 13,  2, 15,  3, 10, 13},
    /* no O's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* P */{  2,  4,  6,  6,  5,  3,  4, 19, 16, 40,  1,  0,  1,  1, 23},
    /* Q */{  0,  0,  0,  2,  2,  1,  0,  1, 13, 12,  3, 25,  0, 30, 15},
    /* R */{  3,  2,  1,  1,  2,  3,  2,  0,  6,  5,  3,  4,  1,  3,  2},
    /* S */{ 16, 21, 19, 30, 20, 17, 29, 72, 39, 53, 25, 18, 11, 12, 16},
    /* T */{ 15, 11, 20, 17, 13, 15, 18, 23, 26, 19, 10, 13,  3,  6, 32},
    /* no U's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* V */{ 28, 32, 21, 21, 16, 18, 18, 11, 16, 14, 33, 14,  2, 12, 17},
    /* W */{  0,  0,  1,  1,  1,  1,  2,  1,  1,  0,  0,  5,  0,  1,  3},
    /* no X's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* Y */{  1,  1,  0,  2,  3,  2,  4,  1,  0, 10,  1,  9,  0,  5,  1},
    /* no Z's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
  },
  /* Gram-positive cleavage sites */
  {
    /*   -13 -12 -11 -10  -9  -8  -7  -6  -5  -4  -3  -2  -1   1   2 */
    /* A */{ 29, 27, 17, 32, 25, 17, 19, 27, 14, 13, 77, 18,114, 58, 15},
    /* no B's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* C */{  2,  2,  2,  1,  1,  1,  1,  0,  0,  0,  0,  1,  0,  0,  0},
    /* D */{  0,  1,  1,  0,  1,  0,  0,  4,  3,  7,  1,  3,  0, 15, 13},
    /* E */{  2,  0,  1,  1,  1,  1,  1,  3,  4,  3,  5, 11,  2, 13, 15},
    /* F */{  8,  8,  6, 11, 14,  5, 12,  1,  1,  3,  2, 10,  0,  2,  0},
    /* G */{  6, 13, 14, 14, 12, 17, 12,  9, 16,  7,  4, 11,  5,  1, 10},
    /* H */{  0,  0,  0,  0,  1,  0,  2,  3,  1,  2,  1,  8,  0,  1,  4},
    /* I */{ 10, 10, 14, 10,  7,  8,  6,  8,  2,  5,  4,  2,  0,  3,  4},
    /* no J's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* K */{  1,  0,  1,  0,  0,  0,  0,  2,  6,  7,  0, 12,  2,  8,  4},
    /* L */{ 26, 31, 26, 27, 31, 32, 18, 10,  5,  5,  3,  5,  0,  1,  1},
    /* M */{  3,  5,  4,  3,  2,  7,  3,  3,  0,  5,  0,  1,  0,  0,  0},
    /* N */{  1,  1,  0,  1,  5,  2,  7,  3, 13,  9,  1,  6,  1,  6,  5},
    /* no O's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* P */{  3,  3,  5,  3,  9,  6, 16, 22, 20, 15,  1,  2,  0,  2, 16},
    /* Q */{  0,  3,  0,  0,  0,  4,  7,  5,  8,  5,  0, 17,  1,  8,  6},
    /* R */{  1,  0,  0,  2,  0,  1,  0,  1,  1,  3,  2,  1,  3,  1,  1},
    /* S */{ 14, 11, 17, 13, 13, 14, 12, 11, 18, 13, 12, 19,  8, 14, 17},
    /* T */{ 17, 13, 14,  7,  7, 12,  9, 16, 19, 28,  1,  4,  1,  4, 21},
    /* no U's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* V */{ 18, 12, 18, 15, 11, 10, 14, 12, 10,  9, 27,  3,  3,  2,  8},
    /* W */{  0,  0,  0,  1,  1,  2,  1,  0,  0,  0,  0,  2,  0,  1,  0},
    /* no X's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
    /* Y */{  0,  1,  1,  0,  0,  2,  1,  1,  0,  2,  0,  5,  1,  1,  1},
    /* no Z's */{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
  }  /* end Gram-positive */
  }; /* end cleavage */

  /** Number of table to use, 0, 1, or 2, for eukaryotic, gram-negative, or
   * gram-positive */
  public int table = -1; 

  /** Easy encoding of 3 chars into an unsigned long, to switch on. 
   * Used to get a unique int from 3 base chars.
   * The chars are passed in as ints to save on conversion time. */
  public static final int e(byte a, byte b, byte c)
  {
    return (((a<<8)+b)<<8)+c;
  }
  public static final int TTT = ((('T'<<8)+'T')<<8)+'T';
  public static final int TTC = ((('T'<<8)+'T')<<8)+'C';
  public static final int TTG = ((('T'<<8)+'T')<<8)+'G';
  public static final int TTA = ((('T'<<8)+'T')<<8)+'A';
  public static final int TCT = ((('T'<<8)+'C')<<8)+'T';
  public static final int TCC = ((('T'<<8)+'C')<<8)+'C';
  public static final int TCG = ((('T'<<8)+'C')<<8)+'G';
  public static final int TCA = ((('T'<<8)+'C')<<8)+'A';
  public static final int TGT = ((('T'<<8)+'G')<<8)+'T';
  public static final int TGC = ((('T'<<8)+'G')<<8)+'C';
  public static final int TGG = ((('T'<<8)+'G')<<8)+'G';
  public static final int TGA = ((('T'<<8)+'G')<<8)+'A';
  public static final int TAT = ((('T'<<8)+'A')<<8)+'T';
  public static final int TAC = ((('T'<<8)+'A')<<8)+'C';
  public static final int TAG = ((('T'<<8)+'A')<<8)+'G';
  public static final int TAA = ((('T'<<8)+'A')<<8)+'A';
  public static final int CTT = ((('C'<<8)+'T')<<8)+'T';
  public static final int CTC = ((('C'<<8)+'T')<<8)+'C';
  public static final int CTG = ((('C'<<8)+'T')<<8)+'G';
  public static final int CTA = ((('C'<<8)+'T')<<8)+'A';
  public static final int CCT = ((('C'<<8)+'C')<<8)+'T';
  public static final int CCC = ((('C'<<8)+'C')<<8)+'C';
  public static final int CCG = ((('C'<<8)+'C')<<8)+'G';
  public static final int CCA = ((('C'<<8)+'C')<<8)+'A';
  public static final int CGT = ((('C'<<8)+'G')<<8)+'T';
  public static final int CGC = ((('C'<<8)+'G')<<8)+'C';
  public static final int CGG = ((('C'<<8)+'G')<<8)+'G';
  public static final int CGA = ((('C'<<8)+'G')<<8)+'A';
  public static final int CAT = ((('C'<<8)+'A')<<8)+'T';
  public static final int CAC = ((('C'<<8)+'A')<<8)+'C';
  public static final int CAG = ((('C'<<8)+'A')<<8)+'G';
  public static final int CAA = ((('C'<<8)+'A')<<8)+'A';
  public static final int GTT = ((('G'<<8)+'T')<<8)+'T';
  public static final int GTC = ((('G'<<8)+'T')<<8)+'C';
  public static final int GTG = ((('G'<<8)+'T')<<8)+'G';
  public static final int GTA = ((('G'<<8)+'T')<<8)+'A';
  public static final int GCT = ((('G'<<8)+'C')<<8)+'T';
  public static final int GCC = ((('G'<<8)+'C')<<8)+'C';
  public static final int GCG = ((('G'<<8)+'C')<<8)+'G';
  public static final int GCA = ((('G'<<8)+'C')<<8)+'A';
  public static final int GGT = ((('G'<<8)+'G')<<8)+'T';
  public static final int GGC = ((('G'<<8)+'G')<<8)+'C';
  public static final int GGG = ((('G'<<8)+'G')<<8)+'G';
  public static final int GGA = ((('G'<<8)+'G')<<8)+'A';
  public static final int GAT = ((('G'<<8)+'A')<<8)+'T';
  public static final int GAC = ((('G'<<8)+'A')<<8)+'C';
  public static final int GAG = ((('G'<<8)+'A')<<8)+'G';
  public static final int GAA = ((('G'<<8)+'A')<<8)+'A';
  public static final int ATT = ((('A'<<8)+'T')<<8)+'T';
  public static final int ATC = ((('A'<<8)+'T')<<8)+'C';
  public static final int ATG = ((('A'<<8)+'T')<<8)+'G';
  public static final int ATA = ((('A'<<8)+'T')<<8)+'A';
  public static final int ACT = ((('A'<<8)+'C')<<8)+'T';
  public static final int ACC = ((('A'<<8)+'C')<<8)+'C';
  public static final int ACG = ((('A'<<8)+'C')<<8)+'G';
  public static final int ACA = ((('A'<<8)+'C')<<8)+'A';
  public static final int AGT = ((('A'<<8)+'G')<<8)+'T';
  public static final int AGC = ((('A'<<8)+'G')<<8)+'C';
  public static final int AGG = ((('A'<<8)+'G')<<8)+'G';
  public static final int AGA = ((('A'<<8)+'G')<<8)+'A';
  public static final int AAT = ((('A'<<8)+'A')<<8)+'T';
  public static final int AAC = ((('A'<<8)+'A')<<8)+'C';
  public static final int AAG = ((('A'<<8)+'A')<<8)+'G';
  public static final int AAA = ((('A'<<8)+'A')<<8)+'A';

  /** Returns 'A' on input 'T', 'C' on input 'G', and vice versa, the
   * "pair" base on the other part of the double helix. Uses ints instead
   * of chars to reduce conversion time. */
  public static byte pair(byte base)
  {
    switch(base){
    case 'A':
      return 'T';
    case 'T':
      return 'A';
    case 'C':
      return 'G';
    case 'G':
      return 'C';
    default:
      throw new IllegalArgumentException("Unknown base " + (char)base 
					 + " (" + base + ")");
    }
  }
  
  /** Returns amino acid character for input base triplet. */
  public static byte lookup(byte[] base)
  {
    int c, i;

    i = (((base[0]<<8)+base[1])<<8)+base[2];
    
    /* printf("looking up '%s' = %lu\n", base, i); */
    switch(i)
    {
    case TTT:case TTC:
      c = 'F'; break;
      
    case TTA:case TTG:    case CTT:case CTC:    case CTA:case CTG:
      c = 'L'; break;
      
    case ATT:case ATC:    case ATA:
      c = 'I'; break;
    case ATG:
      c = 'M'; break;
      
    case GTT:case GTC:    case GTA:case GTG:
      c = 'V'; break;
      
    case TCT:case TCC:    case TCA:case TCG:    case AGT:case AGC:
      c = 'S'; break;
      
    case CCT:case CCC:    case CCA:case CCG:
      c = 'P'; break;
      
    case ACT:case ACC:    case ACA:case ACG:
      c = 'T'; break;
      
    case GCT:case GCC:    case GCA:case GCG:
      c = 'A'; break;
      
    case TAT:case TAC:
      c = 'Y'; break;
    case TAA:case TAG:    case TGA:
      c = '*'; break; /* special "stop codon" */
      
    case CAT:case CAC:      
      c = 'H'; break;
    case CAA:case CAG:
      c = 'Q'; break;
      
    case AAT:case AAC:
      c = 'N'; break;
    case AAA:case AAG:
      c = 'K'; break;
      
    case GAT:case GAC:
      c = 'D'; break;
    case GAA:case GAG:
      c = 'E'; break;
      
    case TGT:case TGC:
      c = 'C'; break;
    case TGG:
      c = 'W'; break;
      
    case CGT:case CGC:    case CGA:case CGG:    case AGA:case AGG:
      c = 'R'; break;
      
    case GGT:case GGC:    case GGA:case GGG:
      c = 'G'; break;
      
    default:   
      throw new IllegalArgumentException("Unknown base " + (char)base[0] 
					 + (char)base[1] + (char)base[2]);

    }
    return (byte)c;
  }

  /** Returns the next legal base character on input stream, ignoring
   * any illegal others. Returns -1 on error, like in.read(). */
  int getBase() throws IOException
  {
    int c;
    
    do {
      c = in.read();
      if(c == -1) break;
      c = Character.toUpperCase((char)c);
    } while(c!='A' && c !='C' && c!= 'T' && c!= 'G');
    return c;
  }

  /** Adds an amino acid to a frame, writes the protein if a stop
   * codon. Uses ints instead of chars to reduce conversion time.
   * @param amin amino acid to write, 0 and '*' are special stop codons
   * @param frame number of frame to write to, 0-5, 3-5 are backwards frames
   * @exception ArrayIndexOutOfBoundsException on buffer overflow */
  void add(byte amin, int frame)
  {
    /*printf("bases %ld: adding char %c to frame %u\n", bases, amin, frame); */
    if(amin == '*' || amin == 0)
    { /* special stop codon */
      int len; /* length of protein string */
      
      if(frame < 3)
      { /* writing forward */
	if(amin != 0)
	  buffer[frame][buffptr[frame]++] = amin;
	len = buffptr[frame];
	writeProtein(frame, bases - len*3, buffer[frame], 
		     0, len);
	buffptr[frame] = 0; /* reset buffptr */
      }
      else
      { /* frame >= 3, writing backward */
	len = MAXPROTEIN - buffptr[frame];
	/* -1 to position because the '*' isn't counted in yet */
	writeProtein(frame, bases - len*3 - 1, buffer[frame],
		     buffptr[frame], MAXPROTEIN);
	buffptr[frame] = MAXPROTEIN; /* reset buffptr */
	if(amin != 0)
	  buffer[frame][--buffptr[frame]] = amin;
      }
    } else { /* write normal, non-stop codon */
      /* Java automatically checks for overflow of buffer */
     if(frame < 3){
	buffer[frame][buffptr[frame]++] = amin;
      } else { /* write backward */
	buffer[frame][--buffptr[frame]] = amin;
      }
    }
  }/* end add() */

  /* Writes the protein to the file.
     <P>
     A typical signal sequence needs to have the positively charged amino
     acids first (K or R or both) which are right next (5 to 10 residues) to
     the start amino acid (M). These positive residues (K,R) are not
     evaluated by the Hejne Matrix. Only the 15 residues which follows next
     to K,or R are evaluated, noted as -13, to 2, and also known as cleavage
     site. Now, if we could let the program first look for the presents of K
     and R right next (window 5 to 10 or variable) to the M and then allow
     the algorithem to start to evaluate the next coming residues according
     to the matrix (15 positions), we definitly would have a much higher
     yield in putative secreted proteins.
     <P>
     First task: i) find an open reading frame defined with the start M and
     check for residues K or R right next (window 1 to 10, optional), then
     ii) apply to such candidates, and only to such, a sliding window (1- 15,
     optional) which compares the following residues to the matrix and stores
     them away if minimal weight factor level is reached.
     <P>
     The first considered (to be compared with the matrix) amino acid is
     position -13 then the secound is -12, down to -1 (here the cut is
     produced) +1 means the mature protein starts.
     <P>
     example:
     <PRE>
          -13         -112
     MFGKKRLALVALVALLLLAXA  (this is a good looking siganl sequence!!!!)
     MFGGDFLALVALVALLLLAXA  (this is not a signal sequence at all, because it
     lacks the positive residues K and/or R!!!!)
     </PRE>
     @param frame index of frame, 0-5, only to write comment header
     @param position start of protein in file, only for comment header
     @param prot array holding protein amino acids
     @param offset beginning of protein in prot array
     @param end one past end of protein (in aminos) in prot array
   */

  void writeProtein(int frame, int position, byte[] prot, int offset, int end)
  {
    int maxWt = 0; /* max weight of protein over all window slides */
    int maxPos=0; /* slide position of max weight of protein */
    int start = offset; /* start of final protein */
       
    /* test for excluded amino acids */
    if(nExcluded > 0)
    {
      for(int i=start; i<end; ++i)
      {
	if(exclude[prot[i] - 'A'])
	{ /* excluded found - skip all such */
	  start = i+1;
	}
      }
    }
    
    /* scan until finding a good enough weight, and acceptable protein */
    for(;;)
    {
      int stop = end-minLength; /* generic counter endpoint */
      /* only start with a methanine */
      while(start < stop && prot[start] !='M')
      {
	start ++;
      }
      if(start >= stop)
	return; /* too short, or no 'M' at all... */
      /* search for positive residues, K or R */
      stop = (start + maxPositiveResidues > end) 
	? end : (start +maxPositiveResidues);
      int endKR=0;/* distance until after last contiguous K/R amino of K/R
		     sequence starting within maxPositiveResidues of start */
      int i; /* generic counter */
      for(i=start; i<stop; ++i)
      {
	if(prot[i] == 'K' || prot[i] == 'R')
	{
	  for(endKR=i+1; prot[endKR] == 'K' || prot[endKR] == 'R'; ++endKR);
	  break;
	}
      }
      if(endKR > 0) /* Positive residue found. */
      {
	stop = (endKR+maxSlide > end) ? end : (endKR+maxSlide);
	/* slide 15 amino acid window along protein residues
	   for first maxSlide amino acids */
	for(i=endKR; i<stop; ++i)
	{
	  int wt = weigh(prot, i, end);
	  if(wt > maxWt)
	  {
	    maxWt = wt;
	    maxPos = i;
	  }
	}
	if(maxWt >= minWeight) break; /* found one */
      }
      /* otherwise reject this whole protein, 
	 try next starting one further down */
      start++;
    }/* end for(;;) - found a good enough weight, now write it  */
  
    if(frame < 3)
      position += 3 * (start - offset);
    /* For the reversed frames the position is the location of the stop
       codon anyway, so where a protein starts doesn't affect that. */

    proteins++; /* wrote another protein */
    out.println("> " + (frame+1) + " / " + (position+1) 
		+ " (" + (end-start) + " amino acids long, weight " + maxWt
		+ " from position " + (maxPos - start + 1) + ")");
    
    /* write one MAXLINE at a time */
    while((end-start) > MAXLINE)
    {
      out.write(prot, start, MAXLINE);
      start += MAXLINE;
      out.println("");
    }
    out.write(prot, start, end-start);
    out.println("");
    out.println(""); /* blank line between proteins */
  }

  /** Weighs the likelihood of the protein being valid, the higher the
   *  better */
  int weigh(byte[] prot, int start, int end)
  {
    int total = 0;
    int len;

    if(prot[end -1] == '*') /* don't weigh the stop codon */
	end --;
    len = end-start;
    if(len > 15) len = 15; /* 15 = width of table */
    for(int i=0; i<len; ++i){
      total += cleavage [table] [ prot[start+i] - 'A' ] [i];
    }
    return total;
  }

  /** Does the work - reads from in stream and writes to out stream. 
   * @exception RuntimeException - on IOException - sorry!*/
  public void run()
  {
    try 
    {
      /* read 2 first bases, to "prime" buffers */
      int c = getBase();
      if(c==-1)
	throw new RuntimeException("Input file ended on first character!");
      fbuf[1] = (byte)c;
      rbuf[1] = pair((byte)c);
  
      c = getBase();
      if(c==-1)
	throw new RuntimeException("Input file ended on second character!");
      fbuf[2] = (byte)c;
      rbuf[0] = pair((byte)c);
  
      bases = 3; /* number of bases read */
  
      /* main loop, reading bases and writing amino acids to buffers */
      for(int frame=0;(c=getBase())!=-1; frame=(frame+1)%3)
      {
	/* shift back buffers */
	fbuf[0] = fbuf[1];
	fbuf[1] = fbuf[2];
	fbuf[2] = (byte)c;
	add(lookup(fbuf), frame);
    
	rbuf[2] = rbuf[1];
	rbuf[1] = rbuf[0];
	rbuf[0] = pair((byte)c);    
	add(lookup(rbuf), frame+3);

	if(++bases%1000 == 0)
	  System.out.println("Read " + bases + " bases - wrote " + proteins 
			     + " proteins.");
      }
      for(int frame=0; frame<6; ++frame)
	add((byte)0, frame);        /* phony stop */
      in.close();
      out.close();
    }
    catch(IOException e)
    {
      e.printStackTrace(out);
      out.flush();
      throw new RuntimeException(e.toString());
    }
  }

  /** Get user input, start point of program. */
  public static void main(String args[]) throws IOException
  {
    try{
    String s;
    DataInputStream in = new DataInputStream(System.in);
    int maxWeight;
    Bas2Prot bas2prot = new Bas2Prot(); 
    
    System.out.println("bas2prot, base to likely protein text file conversion program.");
    System.out.println("George Ruban 04-09-97 for Joachim Reidl.\n");
    
    System.out.print("Name of input file (including path)? ");
    System.out.flush();
    s = in.readLine();
    bas2prot.in = new BufferedInputStream(new FileInputStream(s));
    
    System.out.print("Name of output file (including path)? ");
    System.out.flush();
    s = in.readLine();
    bas2prot.out = new 
      PrintStream(new BufferedOutputStream(new FileOutputStream(s)));

    System.out.print("Cleavage table to be used: eukaryotic(0), gram-negative(1), gram-positive(2)? ");
    System.out.flush();
    s = in.readLine();
    bas2prot.table = Integer.parseInt(s);
    switch(bas2prot.table){
    case 0:
      System.out.println("Using eukaryotic signal sequence table.");
      maxWeight = MAXEUKARYOTIC;
      break;
    case 1:
      System.out.println("Using gram-negative signal sequence table.");
      maxWeight = MAXGRAMNEGATIVE;
      break;
    case 2:
      System.out.println("Using gram-positive signal sequence table.");
      maxWeight = MAXGRAMPOSITIVE;
      break;
    default:
      throw new IllegalArgumentException("Unknown signal sequence table " + 
					 bas2prot.table);
    }

    System.out.print("Minimum weight to write a protein (0 to " + maxWeight +
		     ") ? ");
    System.out.flush();
    s = in.readLine();
    bas2prot.minWeight = Integer.parseInt(s);
    if(bas2prot.minWeight < 0 || bas2prot.minWeight > maxWeight)
      throw new 
	IllegalArgumentException("Illegal weight " + bas2prot.minWeight +
				 ".  Must be between 0 and " + maxWeight);

    System.out.print("Maximum number of times to slide weighing window? ");
    System.out.flush();
    s = in.readLine();
    bas2prot.maxSlide = Integer.parseInt(s);
    if(bas2prot.maxSlide < 1)
      bas2prot.maxSlide = 1; /* '0' shorthand for '1' (?) */

    System.out.print("Maximum distance of positive residue (K or R)\n" +
		     "from protein start codon (M) (1 to 10, or more)? ");
    System.out.flush();
    s = in.readLine();
    bas2prot.maxPositiveResidues = Integer.parseInt(s);
    if(bas2prot.maxPositiveResidues < 1)
      bas2prot.maxPositiveResidues = 10; /* '0' shorthand for '10' (?) */

    System.out.print("Minimum length (in amino acids) to write a protein? ");
    System.out.flush();
    s = in.readLine();
    bas2prot.minLength = Integer.parseInt(s);
    if(bas2prot.minLength < 0 || bas2prot.minLength > MAXPROTEIN)
      throw new 
	IllegalArgumentException("Illegal minimum length " + 
				 bas2prot.minLength +
				 " must be between 0 and " + MAXPROTEIN);
    if(bas2prot.minLength <1)
      bas2prot.minLength = 1; /* '0' shorthand for '1' (?) */
    System.out.print("Any amino acids to be excluded from database "
		       + "(especially W, P, N, R, K, D, or C )? ");
    System.out.flush();
    s = in.readLine();
    for(int i=0; i<s.length(); ++i)
    {
      if(Character.isUpperCase(s.charAt(i)))
      {
	bas2prot.exclude[ s.charAt(i) - 'A'] = true;
	++bas2prot.nExcluded;
      }
      else if(Character.isLowerCase(s.charAt(i)))
      {
	bas2prot.exclude[ s.charAt(i) - 'a'] = true;
	++bas2prot.nExcluded;
      }
    }
    if(bas2prot.nExcluded > 0){
      System.out.println("Excluding " + bas2prot.nExcluded + " amino acids:");
      for(int i = 0; i<bas2prot.exclude.length; ++i)
	if(bas2prot.exclude[i]) System.out.print(" "+ (char)(i + 'A'));
      System.out.println("");
    }
    System.out.println("Beginning program.");
    bas2prot.run();
    System.out.println("Done! " + bas2prot.bases + " bases read, " + 
		       bas2prot.proteins + " proteins written.");
    }catch(Exception e)
      {
	e.printStackTrace();
	System.out.println("Hit Enter to Exit");
	System.in.read();
      }
  }/* end main() */
}
